<?php

namespace LakeDrops\DockerTraefik;

use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Yaml\Yaml;

/**
 * Provides configuration methods for Traefik.
 *
 * @package LakeDrops\DockerTraefik
 */
class Traefik {

  /**
   * The name of the project using Traefik.
   *
   * @var string
   */
  protected string $name;

  /**
   * The domain.
   *
   * @var string
   */
  protected string $domain;

  /**
   * The HTTP port.
   *
   * @var int
   */
  protected int $httpPort;

  /**
   * The HTTPS port.
   *
   * @var int
   */
  protected int $httpsPort;

  /**
   * The certificate filename.
   *
   * @var string
   */
  protected string $certFilename;

  /**
   * The key filename.
   *
   * @var string
   */
  protected string $keyFilename;

  /**
   * Flag for enabling the Portainer addon.
   *
   * @var bool
   */
  protected bool $addonPortainer = FALSE;

  /**
   * The Trafik hub token.
   *
   * @var string
   */
  protected string $hubToken = '';

  /**
   * The environment variables.
   *
   * @var array
   */
  protected array $env;

  /**
   * Flag, if tls should be used.
   *
   * @var bool
   */
  protected bool $tls;

  /**
   * Flag, if the DNS challenge should be used.
   *
   * @var bool
   */
  protected bool $dnsChallenge;

  /**
   * The DNS challenge provider.
   *
   * @var string
   */
  protected string $dnsChallengeProvider;

  /**
   * The DNS challenge resolver.
   *
   * @var string
   */
  protected string $dnsChallengeResolver;

  /**
   * Traefik constructor.
   *
   * @param string $name
   *   Name of the network to be created which needs to match the container
   *   prefix of your project you would like to handle with Traefik.
   * @param string $domain
   *   The domain name used for all local projects.
   * @param int $httpPort
   *   Port for non secure requests.
   * @param int $httpsPort
   *   Port for secure requests.
   * @param string $certFilename
   *   Filename of the SSL certificate.
   * @param string $keyFilename
   *   Filename of the private key for the SSL certificate.
   * @param array $env
   *   A list of environment variables for the Traefik container.
   * @param bool $tls
   *   Whether TLS should be supported.
   * @param bool $dnsChallenge
   *   If TLS is supported, it uses the http challenge by default. Set to TRUE
   *   to use the DNS challenge.
   * @param string $dnsChallengeProvider
   *   If DNS challenge should be used, a provider is required. For a list of
   *   supported providers.
   * @param string $dnsChallengeResolver
   *   In some scenarios, the DNS resolver needs to be defined in order to
   *   prevent local or shadow DNS servers being used.
   *
   * @see https://doc.traefik.io/traefik/https/acme/#providers
   */
  public function __construct(
    string $name,
    string $domain = 'docker.localhost',
    int $httpPort = 8000,
    int $httpsPort = 8443,
    string $certFilename = '',
    string $keyFilename = '',
    array $env = [],
    bool $tls = FALSE,
    bool $dnsChallenge = FALSE,
    string $dnsChallengeProvider = '',
    string $dnsChallengeResolver = ''
  ) {
    $this->name = $name;
    $this->domain = $domain;
    $this->httpPort = $httpPort;
    $this->httpsPort = $httpsPort;
    $this->certFilename = $certFilename;
    $this->keyFilename = $keyFilename;
    $this->env = $env;
    $this->tls = $tls;
    $this->dnsChallenge = $dnsChallenge;
    $this->dnsChallengeProvider = $dnsChallengeProvider;
    $this->dnsChallengeResolver = $dnsChallengeResolver;
  }

  /**
   * Sets the flag for the Portainer addon.
   *
   * @param bool $addonPortainer
   *   The flag.
   */
  public function setAddonPortainer(bool $addonPortainer): void {
    $this->addonPortainer = $addonPortainer;
  }

  /**
   * Sets the hub token.
   *
   * @param string $hubToken
   *   The hub token.
   */
  public function setHubToken(string $hubToken): void {
    $this->hubToken = $hubToken;
  }

  /**
   * Update the Traefik container.
   */
  public function update(): void {
    // Update host wide traefik container.
    $traefikPath = $_SERVER['HOME'] . '/.traefik';
    $traefikCertPath = $_SERVER['HOME'] . '/.traefik/certs';
    $traefikConfigPath = $_SERVER['HOME'] . '/.traefik/configuration';
    $traefikFile = $traefikPath . '/docker-compose.yml';

    $user = getenv('USER');
    $fs = new Filesystem();
    if (!$fs->exists($traefikPath)) {
      $fs->mkdir($traefikPath);
      if ($user) {
        $fs->chown($traefikPath, $user, TRUE);
      }
    }
    if (!$fs->exists($traefikCertPath)) {
      $fs->mkdir($traefikCertPath);
      if ($user) {
        $fs->chown($traefikCertPath, $user, TRUE);
      }
    }
    if (!$fs->exists($traefikConfigPath)) {
      $fs->mkdir($traefikConfigPath);
      if ($user) {
        $fs->chown($traefikConfigPath, $user, TRUE);
      }
    }
    $upd1 = $this->updateFile($traefikConfigPath . '/certificates.toml', $this->defaultCertificatesConfig());
    $upd2 = $this->updateFile($traefikFile, Yaml::dump($this->defaultDockerCompose(), 9, 2));
    if ($upd1 || $upd2) {
      if ($user) {
        if (file_exists($traefikConfigPath . '/certificates.toml')) {
          $fs->chown($traefikConfigPath . '/certificates.toml', $user, TRUE);
        }
        if (file_exists($traefikFile)) {
          $fs->chown($traefikFile, $user, TRUE);
        }
      }
      $cwd = getcwd();
      chdir($traefikPath);
      exec('docker network create traefik-public 2>/dev/null');
      exec('docker compose --project-name traefik up -d');
      chdir($cwd);
    }
  }

  /**
   * Updates or deletes the file.
   *
   * @param string $filename
   *   The file name.
   * @param string $content
   *   The file content. If empty, the file will be deleted, if it exists.
   *
   * @return bool
   *   TRUE, if an update was necessary, FALSE otherwise.
   */
  private function updateFile(string $filename, string $content): bool {
    if ($content === '') {
      if (file_exists($filename)) {
        unlink($filename);
        return TRUE;
      }
      return FALSE;
    }
    if (file_exists($filename) && file_get_contents($filename) === $content) {
      return FALSE;
    }
    file_put_contents($filename, $content);
    return TRUE;
  }

  /**
   * Gets the default config.
   *
   * @return array
   *   The default config.
   */
  private function defaultDockerCompose(): array {
    $config = [
      'version' => '3.3',
      'services' => [
        'traefik' => [
          'container_name' => 'traefik',
          'image' => 'traefik:2.11',
          'command' => [
            '--api=true',
            '--api.dashboard=true',
            '--api.insecure=true',
            '--log.level=warning',
            '--entrypoints.web.address=:80',
            '--providers.file.directory=/configuration/',
            '--providers.file.watch=true',
            '--providers.docker=true',
            '--providers.docker.exposedbydefault=false',
          ],
          'restart' => 'unless-stopped',
          'networks' => [
            'traefik-public',
          ],
          'ports' => [
            $this->httpPort . ':80',
          ],
          'labels' => [
            'traefik.enable=true',
            'traefik.network=traefik-public',
            'traefik.http.routers.traefik.service=api@internal',
            'traefik.http.routers.traefik.rule=Host(`traefik.' . $this->domain . '`)',
          ],
          'volumes' => [
            './certs:/certs/:ro',
            './configuration:/configuration/:ro',
            '/var/run/docker.sock:/var/run/docker.sock:ro',
          ],
        ],
      ],
      'networks' => [
        'traefik-public' => [
          'external' => TRUE,
        ],
      ],
    ];
    if (!empty($this->env)) {
      $config['services']['traefik']['environment'] = $this->env;
    }
    if ($this->addonPortainer) {
      $config['services']['portainer'] = [
        'container_name' => 'portainer',
        'image' => 'portainer/portainer-ce:2.17.1',
        'command' => '-H unix:///var/run/docker.sock',
        'restart' => 'unless-stopped',
        'networks' => [
          'traefik-public',
        ],
        'labels' => [
          'traefik.enable=true',
          'traefik.network=traefik-public',
          'traefik.http.routers.frontend.rule=Host(`portainer.' . $this->domain . '`)',
          'traefik.http.routers.frontend.entrypoints=websecure',
          'traefik.http.services.frontend.loadbalancer.server.port=9000',
          'traefik.http.routers.frontend.service=frontend',
          'traefik.http.routers.edge.rule=Host(`edge.' . $this->domain . '`)',
          'traefik.http.routers.edge.entrypoints=websecure',
          'traefik.http.services.edge.loadbalancer.server.port=8000',
          'traefik.http.routers.edge.service=edge',
        ],
        'volumes' => [
          './portainerdata:/data',
          '/var/run/docker.sock:/var/run/docker.sock:ro',
        ],
      ];
    }
    if ($this->hubToken) {
      $config['services']['traefik']['command'][] = '--experimental.hub=true';
      $config['services']['traefik']['command'][] = '--hub.tls.insecure=true';
      $config['services']['traefik']['command'][] = '--metrics.prometheus.addrouterslabels=true';
      $config['services']['traefik']['ports'][] = '9900:9900';
      $config['services']['traefik']['ports'][] = '9901:9901';
      $config['services']['hub-agent'] = [
        'container_name' => 'hub-agent',
        'image' => 'ghcr.io/traefik/hub-agent-traefik:v1.0.1',
        'command' => [
          'run',
          '--hub.token=' . $this->hubToken,
          '--auth-server.advertise-url=http://hub-agent',
          '--traefik.host=traefik',
          '--traefik.tls.insecure=true',
        ],
        'restart' => 'on-failure',
        'networks' => [
          'traefik-public',
        ],
        'volumes' => [
          '/var/run/docker.sock:/var/run/docker.sock:ro',
        ],
        'depends_on' => [
          'traefik',
        ],
      ];
    }
    if ($this->tls) {
      $config['services']['traefik']['command'][] = '--entrypoints.websecure.address=:443';
      $config['services']['traefik']['command'][] = '--entrypoints.web.http.redirections.entrypoint.to=websecure';
      $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.email=admin@' . $this->domain;
      $config['services']['traefik']['ports'][] = $this->httpsPort . ':443';
      $config['services']['traefik']['labels'][] = 'traefik.http.routers.traefik.tls=true';
      $config['services']['traefik']['labels'][] = 'traefik.http.routers.traefik.tls.certresolver=lakedrops';
      if ($this->addonPortainer) {
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.frontend.tls=true';
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.frontend.tls.certresolver=lakedrops';
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.edge.tls=true';
        $config['services']['portainer']['labels'][] = 'traefik.http.routers.edge.tls.certresolver=lakedrops';
      }
      if ($this->dnsChallenge && $this->dnsChallengeProvider !== '') {
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.dnschallenge=true';
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.dnschallenge.provider=' . $this->dnsChallengeProvider;
        if ($this->dnsChallengeResolver !== '') {
          $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.dnschallenge.resolvers=' . $this->dnsChallengeResolver;
        }
      }
      else {
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.storage=/etc/traefik/acme/acme.json';
        $config['services']['traefik']['command'][] = '--certificatesresolvers.lakedrops.acme.httpchallenge.entrypoint=web';
      }
    }
    return $config;
  }

  /**
   * Get the default certificate config.
   *
   * @return string
   *   The default certificate config.
   */
  private function defaultCertificatesConfig(): string {
    if ($this->certFilename === '' || $this->keyFilename === '') {
      return '';
    }
    return <<<EOF
[[tls.certificates]]
   certFile = "/certs/$this->certFilename"
   keyFile = "/certs/$this->keyFilename"
EOF;

  }

}
